local json = require("json")
local const = require("const")
local utils = require("utils3")
local pl_string = require("pl.stringx")
local resty_string = require("resty.string")
local resty_random = require("resty.random")
local resty_cookie = require("resty.cookie")

local memkv = ngx.shared.memkv

local COOKIE_MAX_AGE = 3600

utils.check_request_method("POST")

-- Extract username and password.
local authorization = ngx.var.http_Authorization
if type(authorization) ~= "string" or #authorization == 0 then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Authorization in header not a string or is empty")
end
local start, _, encoded_auth = string.find(authorization, "Basic%s+(.+)")
if not start then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Authorization in header not valid Basic auth string")
end
local auth = ngx.decode_base64(encoded_auth)
if not auth then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Authorization in header not base64 encoded")
end
local items = pl_string.split(auth, ":")
local username = items[1]
local password = items[2]
local token = resty_string.to_hex(resty_random.bytes(32))

local permissions = ""
local ok, ret, err = utils.get_data(const.data_key.eweb.user)
if ok then
    local users
    ok, users = pcall(json.decode, ret)
    if not ok or type(users) ~= "table" or not users["UserList"] then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.SERVICE_DATA_GET,
                "cannot retrieve users or not a JSON object")
    end

    local user = users["UserList"][username]
    if not user then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "no such user: " .. username)
    end

    local salt = user["PasswordSalt"]
    local password_hash = utils.hmac_sha256(salt, password)
    if not password_hash then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.HMAC,
                "failed to hash password")
    end

    if resty_string.to_hex(password_hash) ~= user["PasswordHash"] then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "username or password incorrect")
    end

    for _, perm in ipairs(user["Permission"]) do
        permissions = permissions .. "," .. perm
    end
    -- Add function compute permission for default user.
    if username == "admin" and not string.find(permissions, "FunctionComputeManage") then
        permissions = permissions .. ",FunctionComputeManage"
    end
    -- Add scene management permission for default user.
    if username == "admin" and not string.find(permissions, "SceneManage") then
        permissions = permissions .. ",SceneManage"
    end

    user["LastLoginTime"] = tostring(os.time())
    local updated_users = json.encode(users)
    local ok, _, err_m = utils.set_data(const.data_key.eweb.user, updated_users)
    if not ok then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.SERVICE_DATA_SET,
                "failed to save users: " .. err_m)
    end
else
    if ret ~= const.SERVICE_DATA_KEY_NOT_EXIST then
        return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, ret, err)
    end
    -- Default username and password.
    if username ~= const.default_user.name or password ~= const.default_user.password then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "username or password incorrect")
    end
    permissions = "GatewayManage"
end

local cookie, err_n = resty_cookie:new()
if not cookie then
    return utils.error_exit(
            ngx.HTTP_INTERNAL_SERVER_ERROR,
            const.COOKIE_HANDLER_CREATE, err_n)
end

-- Set access cookie.
local err_c
ok, err_c = cookie:set({
    key = "token",
    value = token,
    path = "/",
    max_age = COOKIE_MAX_AGE,
})
if not ok then
    return utils.error_exit(
            ngx.HTTP_INTERNAL_SERVER_ERROR,
            const.COOKIE_SET, err_c)
end

memkv:set(token, permissions, COOKIE_MAX_AGE)
ngx.status = ngx.HTTP_CREATED
ngx.header['Content-Type'] = 'application/json; charset=utf-8'
ngx.exit(ngx.HTTP_OK)