local const = require("const")
local json = require("json")
local utils = require("utils3")

utils.check_request_method("POST")
utils.check_auth()
utils.check_permission("DeviceManage")

local function import_things(things)
    local params = {}
    for k, v in pairs(things) do
        local i = string.find(k, "_")
        local product = string.sub(k, 1, i-1)
        local device = string.sub(k, i+1)
        table.insert(params, { product, device, v })
    end

    local req = { params = params }
    local ret, err, cause = utils.dbus_call_edge(
            const.dimu.wkn, const.dimu.obj, const.dimu.inf,
            "importDeviceTrinity", json.encode(req))
    if not ret then
        return false, err, cause
    end

    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, const.MESSAGE_INVALIDREPLY, "not a JSON object"
    end
    if res["code"] ~= 0 then
        local msg = { code = res["code"], message = res["message"] }
        return false, const.MESSAGE_FAILEDREPLY, json.encode(msg)
    end
    return true
end

local function get_things()
    local ok, data, err = utils.get_data("__DIMU_DEVICE_TRINITY")
    if not ok then
        if data ~= const.SERVICE_DATA_KEY_NOT_EXIST then
            return false, data, err
        end
        -- Update the returned data if key does not exists.
        data = "{}"
    end
    local ok, things = pcall(json.decode, data)
    if not ok or type(things) ~= "table" then
        return false, const.const.SERVICE_DATA_GET, "not a JSON object"
    end
    return true, things
end

ngx.req.read_body()
local body = ngx.req.get_body_data()
if not body then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "no body found")
end

local ok, args = pcall(json.decode, body)
if not ok then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "not a JSON body")
end

local things = args["Things"]
if not utils.is_array_table(things) or #things == 0 then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Things not a array or is empty")
end

local ok, exists, err = get_things()
if not ok then
    return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, exists, err)
end

local updated_things = exists
for _, v in pairs(things) do
    -- Handle parameters.
    local product = v["ProductKey"]
    if type(product) ~= "string" or #product == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "ProductKey not a string or is empty")
    end
    if not utils.is_valid_pk(product) then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "illegal ProductKey")
    end

    local device = v["DeviceName"]
    if type(device) ~= "string" or #device == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "DeviceName not a string or is empty")
    end
    if not utils.is_valid_dn(device) then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "illegal DeviceName")
    end

    local secret = v["DeviceSecret"]
    if type(secret) ~= "string" or #secret == 0 then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "DeviceSecret not a string or is empty")
    end
    if not utils.is_valid_ds(secret) then
        return utils.error_exit(
                ngx.HTTP_BAD_REQUEST,
                const.INVALID_PARAMETER,
                "illegal DeviceSecret")
    end
    local product_device = product .. "_" .. device
    updated_things[product_device] = secret
end
local ok, err, cause = import_things(updated_things)
if not ok then
    return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, err, cause)
end

ngx.header['Content-Type'] = 'application/json; charset=utf-8'
ngx.exit(ngx.HTTP_OK)