local const = require("const")
local json = require("json")
local utils = require("utils")

utils.check_login()
utils.check_permission("DeviceManage")

local request_method = ngx.var.request_method
if request_method ~= "POST" then
    return utils.error_exit(405, "request method not allowed: " .. request_method)
end

local function list_devices()
    local ok, ret, err = utils.dbus_call_edge(
            const.dimu.wkn, const.dimu.obj, const.dimu.inf,
            "getDeviceState")
    if not ok then
        return false, nil, err
    end

    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, nil, "not a JSON object retrieved from device manager"
    end
    if res["code"] ~= 0 then
        return false, nil, tostring(res["message"])
    end
    if not res["params"] or type(res["params"]) ~= "table" or
            not res["params"]["devices"] or type(res["params"]["devices"]) ~= "table" then
        return false, nil, "invalid return value from device manager"
    end

    local ok, gw_product, gw_device = utils.get_gw_config_2()
    if not ok then
        utils.log_info("failed to retrieve gateway triples: " .. gw_product)
        gw_product = ""
        gw_device = ""
    end

    local devices = {}
    for _, v in ipairs(res["params"]["devices"]) do
        local product_key = v["productKey"]
        local device_name = v["deviceName"]
        local device = {}
        device["ProductName"] = ""
        device["ProductKey"] = product_key or ""
        device["DeviceName"] = device_name or ""
        device["NickName"] = v["nickName"] or ""
        device["DriverId"] = v["driverId"] or ""
        device["DriverName"] = v["driverName"] or ""
        device["Tags"] = v["tags"] or {}
        device["Status"] = "Inactivated"
        device["ConnectTime"] = ""
        device["Connected"] = false

        if product_key then
            local ok, tsl_str = utils.get_tsl(product_key)
            if ok then
                local tsl
                ok, tsl = pcall(json.decode, tsl_str)
                if ok then
                    if tsl and tsl["profile"] then
                        device["ProductName"] = tsl["profile"]["productName"] or ""
                    end
                else
                    utils.log_info("decode tsl %s: %s", tsl_str, tsl)
                end
            else
                utils.log_info("get tsl for %s: %s", product_key, tsl_str)
            end

            -- If the device is the gateway.
            if device_name and device_name == gw_device and product_key == gw_product then
                device["IsGateway"] = true
            end
        end

        if v["localOnline"] == "True" then
            device["Connected"] = true
        end

        local status
        if v["activation"] == "activated" then
            if v["cloudOnline"] == "True" then
                status = "Online"
            else
                status = "Offline"
            end
        elseif v["activation"] == "activationFailed" then
            status = "Failed"
        end
        if status then
            device["Status"] = status
        end

        if v["device"][3] then
            device["ConnectTime"] = v["lastLocalOnlineTime"] or 0
        end
        table.insert(devices, device)
    end
    return true, devices, nil
end

local utils_header = require("utils_header")
utils_header.check_public_body_data()

local ok, devices, err = list_devices()
if not ok then
    return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, err)
end

local ret_data = {}
ret_data["List"] = devices
utils.error_exit(200, json.encode(ret_data))